package net.sf.sql2java.common.beans;

import java.io.Serializable;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * Processable.java
 *     Created on 22-jul-2007, 11:24:28
 * 
 * @author dge2
 */
public class Processable 
	extends HashMap<String, Object> 
	implements Serializable, Cloneable 
{
	private static final long serialVersionUID = -2600978031623122929L;
	
	public static final String PARAMETER_ID 	= "id";
	public static final String PARAMETER_NAME 	= "name";
	public static final String PARAMETER_KEY 	= "key";
	public static final String PARAMETER_SCOPE 	= "scope";
	public static final String PARAMETER_PARENT = "parent";
	
	private Map<String, Map<String, Processable>> references;
	
// -- Constructors ------------------------------------------------------------
	
	public Processable(String id, String name, Processable parent, String scope) {
		put(PARAMETER_ID, id);
		put(PARAMETER_NAME, name);
		put(PARAMETER_SCOPE, scope);
		put(PARAMETER_PARENT, parent);
		
		// -- create the key
		put(PARAMETER_KEY, createUniqueKey(parent, scope, name));
		
		references = new HashMap<String, Map<String, Processable>>();
	}
	
	private Processable() { 
		references = new HashMap<String, Map<String, Processable>>();
	}
	
// -- Cloning -----------------------------------------------------------------
	
	@SuppressWarnings("unchecked")
	public Processable clone() {
		// -- create the result
		Processable result = new Processable();
		
		// -- put all the parameters into the new object
		result.putAll(this);
		
		// -- put all references into the new object
		result.references = (Map<String, Map<String, Processable>>) super.clone();
		
		// -- return the result
		return result;
	}
	
// -- Conversion --------------------------------------------------------------
	
	@Override
	public String toString() {
		StringBuffer result = new StringBuffer();

		if (getParent() != null)
			result.append(getParent().toString()).append(".");
		
		result.append(getName());
		
		return result.toString();
	}
	
// -- Unique Key --------------------------------------------------------------
	
    public String getKey() {
    	if (! containsKey(PARAMETER_KEY)) return null;
    	else return (String) get(PARAMETER_KEY);
	}
    
    private void setKey(String key) {
    	put(PARAMETER_KEY, key);
    }
    
    protected String createUniqueKey(Processable parent, String scope, String name) {
    	// -- create the result
    	StringBuffer result = new StringBuffer();
    	
    	// -- get the parentKey
    	if (parent != null) result.append(parent.getKey());
    	else result.append("root");
    	
    	result.append(".").append(scope.toLowerCase());
    	result.append("(").append(name).append(")");
    	
    	// -- return the result
    	return result.toString();
    }
    
// -- Name --------------------------------------------------------------------
    
    public String getName() {
    	if (! containsKey(PARAMETER_NAME)) return null;
    	else return (String) get(PARAMETER_NAME);
	}
    
// -- Parent ------------------------------------------------------------------
    
    public Processable getParent() {
    	if (! containsKey(PARAMETER_PARENT)) return null;
    	else return (Processable) get(PARAMETER_PARENT);
    }
    
    public void setParent(Processable parent) {
		put(PARAMETER_PARENT, parent);
		
		// -- recreate the key
		setKey(createUniqueKey(parent, getScope(), getName()));
	}
    
// -- String -------------------------------------------------------------------
    
    public String getScope() {
    	if (! containsKey(PARAMETER_SCOPE)) return null;
    	else return (String) get(PARAMETER_SCOPE);
	}
    
    public void setScope(String String) {
    	put(PARAMETER_SCOPE, String);
		
		// -- recreate the key
		setKey(createUniqueKey(getParent(), String, getName()));
	}
	
// -- Id ----------------------------------------------------------------------
    
    public String getId() {
    	if (! containsKey(PARAMETER_ID)) return null;
    	else return (String) get(PARAMETER_ID);
	}
    
// -- Reference Manipulation --------------------------------------------------    
    
    /**
     * Add the given processable as a reference.
     * 
     * The String will be used as the parameter key. The value will be a map
     * holding all the processable's for the given String. This way we'll be
     * able to retrieve all the processable's in a given String.
     * 
     * @param reference		The processable to add
     */
	public void addReference(Processable reference) {
    	// -- create the list if needed
    	if (! containsKey(reference.getScope()))
    		references.put(reference.getScope(), new HashMap<String, Processable>());
    	
    	// -- set this processable as the reference's parent
    	reference.setParent(this);
    	
    	// -- add the reference to the String list
    	references.get(reference.getScope())
    			  .put(reference.getName() ,reference);
    }
    
	public void removeReference(Processable reference) {
    	// -- return if String does not exist
    	if (! references.containsKey(reference.getScope())) return;
    	
    	// -- remove the reference
    	references.get(reference.getScope())
    			  .remove(reference.getName());
    	
    	// -- remove the list if there are no items inside
    	if (references.get(reference.getScope()).size() == 0)
    		references.remove(reference.getScope());
    }
    
    /**
     * Get all the references for the given String
     * <p>
     * The map will contain the reference's name (not the reference's key!) 
     * as the key. 
     * 
     * @param String
     * @return
     */
	public Map<String, Processable> getReferences(String String) {
    	// -- create the result
    	Map<String, Processable> result = new HashMap<String, Processable>();
    	
    	if (references.containsKey(String))
    		result = references.get(String);
    	
    	// -- return the result
    	return result;
    }
    
// -- Parameter Manipulation --------------------------------------------------
    
    public void setParameter(String key, Object value) {
    	put(key, value);
    }
    
    public void removeParameter(String key) {
    	remove(key);
    }
        
    @SuppressWarnings("unchecked")
	public <T> T getParameter(String key) {
    	String[] parts = key.split("#");
    	
    	return (T) getParameter(parts);
    }
    
    @SuppressWarnings("unchecked")
	public <T> T getParameter(String[] parts) {
    	// -- return 'null' if the first part wasn't found
    	if (! containsKey(parts[0])) return null;
    	
    	int idxBrace = parts[0].indexOf("(");
    	
    	// -- create the value holder
    	Object value = null;
    	
    	// -- check if we have a collection to deal with
    	if (idxBrace != -1) {
    		String key = parts[0].substring(0, idxBrace);
    		String itemKey = parts[0].substring(
    				idxBrace + 1, parts[0].length() - 2
    		);
    		
    		Map collection = (Map) get(key);
    		
    		if (collection.containsKey(itemKey))
    			value = collection.get(itemKey);
    	} else {
    		value = get(parts[0]);
    	}
    	 
    	if (value instanceof Processable) {
    		String[] toPass = Arrays.copyOfRange(parts, 1, parts.length);
    		
    		value = ((Processable) value).getParameter(toPass);
    	}
    	
    	// -- return the result
    	return (T) value;
    }
    
    public boolean hasParameter(String key) {
    	return containsKey(key);
    }
    
    public Map<String, Object> getParameters() {
    	return this;
    }
    
}
